/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) Original Authors
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

\*---------------------------------------------------------------------------*/

#include "dynamicWhiplashFvMesh.H"
#include "addToRunTimeSelectionTable.H"
#include "volFields.H"
#include "mathematicalConstants.H"

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
    defineTypeNameAndDebug(dynamicWhiplashFvMesh, 0);
    addToRunTimeSelectionTable(dynamicFvMesh, dynamicWhiplashFvMesh, IOobject);
}


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::dynamicWhiplashFvMesh::dynamicWhiplashFvMesh(const IOobject& io)
:
    dynamicFvMesh(io),
    dynamicMeshCoeffs_
    (
        IOdictionary
        (
            IOobject
            (
                "dynamicMeshDict",
                io.time().constant(),
                *this,
                IOobject::MUST_READ,
                IOobject::NO_WRITE
            )
        ).subDict(typeName + "Coeffs")
    ),
    newPoints_
    (
        IOobject
        (
            "points",
            io.time().constant(),
            meshSubDir,
            *this,
            IOobject::MUST_READ,
            IOobject::NO_WRITE
        )
    )
{
    Info<< "Performing a whiplash dynamic mesh calculation" << endl;

    FILE *file_1;
    FILE *file_2;

    //file_1=fopen("constant/vessel_motion_renew.mat", "r");
    //file_2=fopen("constant/vessel_node_y_renew.mat", "r");
    file_1=fopen("constant/vessel_motion.mat", "r");
    file_2=fopen("constant/vessel_node_y.mat", "r");

    Info<< "\nReading constant/vessel_motion_renew.mat:" << endl;
    for(int m=0;m<255;++m){
      for(int v=0;v<6;++v){
        fscanf(file_1, "%f",&P_[m][v]);
        Info<< P_[m][v] << " ";
      }
      Info<<endl;
    }

    fclose(file_1);

    Info<< "\nReading constant/vessel_node_y_renew.mat:" << endl;
    for(int m=0;m<225;++m){
      for(int v=0;v<7;++v){
        fscanf(file_2, "%f",&Y_[m][v]);
        Info<< Y_[m][v] << " ";
      }
      Info<<endl;
    }
    Info<<endl;

    fclose(file_2);

}

// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

Foam::dynamicWhiplashFvMesh::~dynamicWhiplashFvMesh()
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

bool Foam::dynamicWhiplashFvMesh::update()
{

    float x, y, z;
    float delta_1=0.0, delta_2=0.0, delta_x=0.0;
    float J=0.0, K=0.0;
    float up_nod0,dw_nod0,up_nod1,dw_nod1; //0=last timestep; 1=current timestep
    int update_y,section_id;
    //int N=time().value()/time().deltaT().value();
    int N=1+time().value()/0.001;

    Info<< "Integer time:" << N << endl;

    forAll(newPoints_, nIter)
    {
        update_y = 0; 
        section_id = -1;
        z = newPoints_[nIter].component(vector::Z);
        y = newPoints_[nIter].component(vector::Y);
        x = newPoints_[nIter].component(vector::X);

        if (y<0 && update_y==0) //deal with y<0 section
        {
	    //Info<< "Section_id = 0" <<endl;
            update_y = 1;
            section_id = 0;
            J = 0;
            delta_x = x - (P_[N-1][0] * pow(y,5)+
                           P_[N-1][1] * pow(y,4)+
                           P_[N-1][2] * pow(y,3)+
                           P_[N-1][3] * pow(y,2)+
                           P_[N-1][4] * pow(y,1)+
                           P_[N-1][5]);
        }
        for(int i=1;i<=6;++i)
            //N is time step, i is node index and section index
        {
            up_nod0 = Y_[N-1][i];   //last time step
            dw_nod0 = Y_[N-1][i-1];
            up_nod1 = Y_[N][i];     //current time step
            dw_nod1 = Y_[N][i-1];

            if(y>=dw_nod0 && y<=up_nod0 && update_y==0) //IN section 1~6
            { 
	        //Info<< "Section_id = " << i <<endl;
                update_y = 1;
                section_id = i;
                delta_1 = up_nod0-dw_nod0;
                delta_2 = up_nod1-dw_nod1;
                delta_x = x - (P_[N-1][0] * pow(y,5)+
                               P_[N-1][1] * pow(y,4)+
                               P_[N-1][2] * pow(y,3)+
                               P_[N-1][3] * pow(y,2)+
                               P_[N-1][4] * pow(y,1)+
                               P_[N-1][5]);
                J = (y-dw_nod0)*delta_2/delta_1+dw_nod1;
                break;
          }
        }
        if(update_y==0) //IN higher then section 6 part
        {
	    //Info<< "Section_id = 7" <<endl;
            update_y = 1;
            section_id = 7;
            J = Y_[N][6];
            delta_x = x - (P_[N-1][0] * pow(y,5)+
                           P_[N-1][1] * pow(y,4)+
                           P_[N-1][2] * pow(y,3)+
                           P_[N-1][3] * pow(y,2)+
                           P_[N-1][4] * pow(y,1)+
                           P_[N-1][5]);
        }
        if(update_y)
        {
            //Calculate X  value from motion equation:
            K = P_[N][0] * pow(J,5)+
                P_[N][1] * pow(J,4)+
                P_[N][2] * pow(J,3)+
                P_[N][3] * pow(J,2)+
                P_[N][4] * pow(J,1)+
                P_[N][5];
            newPoints_[nIter].component(vector::Z) = z;
            newPoints_[nIter].component(vector::Y) = J;
            newPoints_[nIter].component(vector::X) = delta_x+K;
        }
        else{
          Info<< "Found an un-updated point: " << x << y <<endl;
          //Add fatalError break!
          //Message("find a un-updated point: (%g,%g)",x,y);
          //Error("un-updated point error!");
        }
    }

    fvMesh::movePoints(newPoints_);

    volVectorField& U = 
        const_cast<volVectorField&>(lookupObject<volVectorField>("U"));
    U.correctBoundaryConditions();

    return true;
}

// ************************************************************************* //
